!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   rs_convert_mod
!> @brief   convert restart files
!! @authors Norio Takase (NT), Chigusa Kobayashi (CK)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../../config.h"
#endif

module rs_convert_mod

  use fileio_namd_xyz_mod
  use fileio_namd_xsc_mod
  use rs_option_mod
  use rs_option_str_mod
  use output_str_mod
  use input_str_mod
  use fileio_rst_mod
  use fileio_mod
  use messages_mod
  use constants_mod
 
  implicit none
  private

  ! subroutines
  public  :: convert
  private :: convert_namd
  private :: output_namd

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    convert
  !> @brief        convert restart file
  !! @authors      NT
  !! @param[in]    input  : input information
  !! @param[in]    option : option information
  !! @param[in]    output : output information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine convert(input, option, output)

    ! formal arguments
    type(s_input),           intent(in)    :: input
    type(s_option),          intent(in)    :: option
    type(s_output),          intent(in)    :: output

    ! local variables
    type(s_rst)              :: rst
    type(s_namd_xyz)         :: crd, vel
    type(s_namd_xsc)         :: xsc

    ! check check-only
    if (option%check_only) &
      return

    ! open Source restart file and convert
    select case (option%rstin_format)

    case (RstFormatNAMD)
      call convert_namd(input, option, rst)

    case (RstFormatGENESIS)
      call convert_genesis(input, option, crd, vel, xsc)

    end select

    ! write restart file
    select case (option%rstout_format)

    case (RstFormatNAMD)
      call output_namd(output%coorfile, output%velfile,  output%xscfile,    &
                       crd, vel, xsc)

    case (RstFormatGENESIS)
      call output_rst(output%rstfile, rst)

    end select

    ! deallocate memory
    call dealloc_rst_all(rst)
    call dealloc_namd_xyz(crd)
    call dealloc_namd_xyz(vel)

    return

  end subroutine convert

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    convert_namd
  !> @brief        convert NAMD restart file
  !! @authors      NT
  !! @param[in]    input  : input information
  !! @param[in]    option : option information
  !! @param[inout] rst    : Genesis restart information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine convert_namd(input, option, rst)

    ! parameters
    character(*),  parameter :: RstFileTypes(3) = (/'Min ','MD  ','Remd'/)

    ! formal arguments
    type(s_input),           intent(in)    :: input
    type(s_option),          intent(in)    :: option
    type(s_rst),             intent(inout) :: rst

    !local variables
    type(s_namd_xyz)         :: crd, vel
    type(s_namd_xsc)         :: xsc


    ! input NAMD restart informations
    call input_namd_xyz(input%coorfile, crd)
    if (len_trim(input%velfile) > 0) &
    call input_namd_xyz(input%velfile, vel)
    call input_namd_xsc(input%xscfile, xsc)


    ! rstfile type
    select case(option%rstout_type)

    case (RstTypeAuto)

      if (.not. allocated(vel%v)) then
        rst%rstfile_type = RstfileTypeMin
      else
        rst%rstfile_type = RstfileTypeMd
      end if

    case (RstTypeMin)

      rst%rstfile_type = RstfileTypeMin

    case (RstTypeMD)

      rst%rstfile_type = RstfileTypeMd

    end select

    write(MsgOut,'(A)') 'Run_Namd_Rstcnv> Genesis restart file type: '// &
         RstFileTypes(rst%rstfile_type)

    ! check coordinates and velocities
    if (rst%rstfile_type == rstfileTypeMd) then

      if (.not. allocated(vel%v)) &
        call error_msg('Run_Namd_Rstcnv> velocity file is not indicated.')

      if (size(crd%v) /= size(vel%v)) &
        call error_msg('Run_Namd_Rstcnv> velocity atom count is different'//&
        ' from coordinates.')

    end if

    rst%num_atoms = size(crd%v(1,:))
    call alloc_rst(rst, RestartAtom, rst%num_atoms)


    ! box size
    rst%box_size_x = xsc%a(1)
    rst%box_size_y = xsc%b(2)
    rst%box_size_z = xsc%c(3)

    write(MsgOut,'(A,3F8.3)')  'Run_Namd_Rstcnv> box size : ', &
                               rst%box_size_x, &
                               rst%box_size_y, &
                               rst%box_size_z

    ! min: energy, delta_r
    rst%energy  = option%min_energy
    rst%delta_r = option%min_delta_r

    if (rst%rstfile_type == rstfileTypeMin) then
      write(MsgOut,'(A,F8.3)') 'Run_Namd_Rstcnv> energy   : ', rst%energy
      write(MsgOut,'(A,F8.3)') 'Run_Namd_Rstcnv> delta_r  : ', rst%delta_r
    end if

    ! md : iseed, ndof, thermo_moment, baro_moment
    rst%iseed               = option%md_iseed
    rst%thermostat_momentum = option%md_thermo_moment
    rst%barostat_momentum   = option%md_baro_moment

    if (rst%rstfile_type == rstfileTypeMd) then
      write(MsgOut,'(A,I16)')   'Run_Namd_Rstcnv> iseed    : ', rst%iseed
      write(MsgOut,'(A,F16.9)') 'Run_Namd_Rstcnv> thermostat-momentum : ', &
                               rst%thermostat_momentum
      write(MsgOut,'(A,F16.9)')'Run_Namd_Rstcnv> barostat-momentum   : ', &
                               rst%barostat_momentum(1)
      write(MsgOut,'(A,F16.9)')'                                       ', &
                               rst%barostat_momentum(2)
      write(MsgOut,'(A,F16.9)')'                                       ', &
                               rst%barostat_momentum(3)
    end if

    ! coordinate
    rst%coord(:,:) = crd%v(:,:)

    ! velocity
    if (rst%rstfile_type == rstfileTypeMd) &
      rst%velocity(:,:) = vel%v(:,:)

    write(MsgOut,'(A)') ' '


    ! deallocate memory
    call dealloc_namd_xyz(vel)
    call dealloc_namd_xyz(crd)

    return

900 call error_msg('Run_Rstcnv> Input file open error.')

  end subroutine convert_namd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    convert_genesis
  !> @brief        convert Genesis restart file
  !! @authors      CK
  !! @param[in]    input  : input information
  !! @param[in]    option : option information
  !! @param[in]    coor     : coordinates information
  !! @param[in]    vel      : velocities information
  !! @param[in]    xsc      : extended system information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine convert_genesis(input, option, crd, vel, xsc)

    ! formal arguments
    type(s_input),           intent(in)    :: input
    type(s_option),          intent(in)    :: option
    type(s_namd_xyz),        intent(inout) :: crd, vel
    type(s_namd_xsc),        intent(inout) :: xsc

    !local variables
    type(s_rst)              :: rst


    ! input GENESIS restart informations
    call input_rst(input%rstfile, rst)
    if (rst%rstfile_type == RstfileTypeRemd)  &
       call error_msg('Convert_Genesis> Remd type restart is not allowed')

    ! initialize coor, vel, xsc
    call alloc_namd_xyz(crd, rst%num_atoms)
    call alloc_namd_xyz(vel, rst%num_atoms)
    xsc%step = option%step
    xsc%a(:) = 0.0_wp
    xsc%b(:) = 0.0_wp
    xsc%c(:) = 0.0_wp
    xsc%o(:) = 0.0_wp
    xsc%s(:) = 0.0_wp

    crd%v(:,:) = 0.0_wp
    vel%v(:,:) = 0.0_wp

    ! box size
    xsc%a(1) = rst%box_size_x
    xsc%b(2) = rst%box_size_y
    xsc%c(3) = rst%box_size_z
    xsc%o(1) = option%origin_x
    xsc%o(2) = option%origin_y
    xsc%o(3) = option%origin_z

    write(MsgOut,'(A,3F8.3)')  'Run_Namd_Rstcnv> box size : ', &
                               xsc%a(1), &
                               xsc%b(2), &
                               xsc%c(3)
    write(MsgOut,'(A,3F8.3)')  'Run_Namd_Rstcnv> cell origin : ', &
                               xsc%o(1:3)

    ! coordinate
    crd%v(:,:) = rst%coord(:,:)

    ! velocity
    vel%v(:,:) = rst%velocity(:,:) 

    write(MsgOut,'(A)') ' '

    ! deallocate memory
    call dealloc_rst_all(rst)

    return

900 call error_msg('Run_Rstcnv> Input file open error.')

  end subroutine convert_genesis

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_namd
  !> @brief        a driver subroutine for writing NAMD restart file
  !! @authors      CK
  !! @param[in]    coorfile : filename of coorfile
  !! @param[in]    velfile  : filename of velocity file
  !! @param[in]    xscfile  : extended system information file
  !! @param[in]    coor     : coordinates information
  !! @param[in]    vel      : velocities information
  !! @param[in]    xsc      : extended system information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_namd(coorfile, velfile, xscfile, crd, vel, xsc)

    ! formal arguments
    character(*),            intent(in) :: coorfile
    character(*),            intent(in) :: velfile
    character(*),            intent(in) :: xscfile
    type(s_namd_xyz),        intent(in) :: crd, vel
    type(s_namd_xsc),        intent(in) :: xsc

    ! open xscfile
    !
    call output_namd_xsc(xscfile, xsc)

    ! open coorfile
    !
    call output_namd_xyz(coorfile, crd)

    ! open velfile
    !
    call output_namd_xyz(velfile, vel)

     return

  end subroutine output_namd

end module rs_convert_mod
